// Copyright 2006 FoxyTunes
// All Rights Reserved

// Countries Const:
const USA_COUNTRY = 'USA';
const FRANCE_COUNTRY = 'France';
const CANADA_COUNTRY = 'Canada';
const BRITAIN_COUNTRY = 'Britain';
const GERMANY_COUNTRY = 'Germany';
const JAPAN_COUNTRY = 'Japan';
const AUSTRIA_COUNRTY = 'Austria';

// Class: AlbumCoverGetter
function AlbumCoverGetter(track, album, artist, userCountry, coverSize, handler) {
    this._track = track;
    this._album = album;
    this._artist = artist;

    this._userCountry = userCountry;
    this._handler = handler;
    this._requestObject = null;
    this._imageRetrievalChannel = null;
    this._isAbortCalled = false;
    this._cacheManager = null;
    this._daysToExpire = 30;
    this._daysToExpireForNotFoundItem = 3;
    this._language = null;
    this._tlds = [];
    this._currentTld = null;
    this._amazonAlbumTitle = null;

    switch (coverSize.toLowerCase()) {
		case 'large':
			this._coverSize = 'Large';
			break;

		case 'medium':
			this._coverSize = 'Medium';
			break;

	    case 'small':
		default:
			this._coverSize = 'Small';
			break;
	}

    

    this._amazonKey = '1Q7APN7AZERWDTD5D0R2';
    this._amazonTagArr = new Array();
    this._amazonTagArr['com'] = 'foxytunes-20';
    this._amazonTagArr['ca'] = 'foxytunes0f-20';
    this._amazonTagArr['de'] = 'foxytunes05-21';
    this._amazonTagArr['fr'] = 'foxytunes00-21';
    this._amazonTagArr['co.uk'] = 'foxytunes-21';
    this._amazonTagArr['co.jp'] = 'foxytunes-22';

//    this._amazonTag = 'foxytunes-20';

  
    this._amazonURL = "http://webservices.amazon.%tld%/onca/xml?Service=AWSECommerceService&AWSAccessKeyId=%amazon_key%&AssociateTag=%amazon_tag%&Operation=ItemSearch&SearchIndex=Music&Artist=%artist%%album-query%&ResponseGroup=Small,Images%additional-response-group%";
}

AlbumCoverGetter.scriptsLoaded = false;

AlbumCoverGetter.prototype = {
    loadScripts : function()  {
      if (AlbumCoverGetter.scriptsLoaded) return;

      var scriptLoader =  Components.classes["@mozilla.org/moz/jssubscript-loader;1"].
                              getService(Components.interfaces.mozIJSSubScriptLoader);
      if (scriptLoader) { 
         scriptLoader.loadSubScript("chrome://foxytunes/content/cover-getter/albumCoverCacheManager.js");
      }

      foxytunesInclude(utils_language_detector);
      AlbumCoverGetter.scriptsLoaded = true;
    },

    init: function() {
        this.loadScripts();

	    this._cacheManager = new AlbumCoverCacheManager();
    	try {
	        this._cacheManager.init();
    	} catch (e) {
	        this.printLog("Failed to init cache: " + e);
    	    this._cacheManager = null;
    	}

        if (!this._album) {
	        this._album = ""
    	}
        if (!this._track) {
	        this._track = ""
    	}
        // if there is no artist
        if (!this._artist) {
            this.printLog("No artist was given");
            return false;
        }
        // if there are no album and track
    	if (!this._album  && !this._track) {
	        this.printLog("No album and no track were given");
	        return false;
    	}

	    var languageDetector = new LanguageDetector();
    	this._language = languageDetector.detectLanguage(this._album + this._artist + this._track);
    	this.setTlds();

    	if (this._tlds.length == 0) {
    	    return false;
    	}	

	    this._isAbortCalled = false;

    	return true;
    },
 

    get: function(size) {
        if (typeof size != 'undefined') {
            this._coverSize = size;
        }
        var me = this;

        var albumCoverData = this.getCoverFromCache(this._coverSize);

        if (albumCoverData) {
	    if (albumCoverData.imagePath == null) {
		// if it is a dummy cache item (item that we failed to find recently)
		albumCoverData = null;
	    } 
           // async return
            setTimeout(function() { me.callResponseHandler(albumCoverData); }, 0);
	        return;
    	}
        
        this.getNextAmazonResult();

    },

    
    getNextAmazonResult : function() {
    	if (this._tlds.length == 0) {
            // cacheNotFoundResult will call callResponseHandler after saving in cache
    	    this.cacheNotFoundResult();
    	    return;
    	}   

        this._currentTld = this._tlds.shift();
        //alert(this._currentTld);
        
    	var url = this.generateUrl(this._currentTld);
    	var me = this;
    	this._requestObject = new XMLHttpRequest();
        this.trace(url);
    	this._requestObject.open('get', url, true);
    	this._requestObject.onreadystatechange = function() {
    		if (me._requestObject.readyState == 4) {
		    // for bug in mozilla that throw error when calling status
		    var status =  -1;
		    try {
			status = me._requestObject.status;
		    } catch (e) {
			// print to log?
		    }
   		    if (status == 200) {
    	  	  // Check if not aborted
  	    	  if (Components.isSuccessCode(me._requestObject.channel.status)) {
			    var albumCoverData = me.parseResponse();
			    if (albumCoverData != null) {
    				try {
	    			    if (!me._isAbortCalled) {
		    		        // get the image for cache, call the handler when it arrives
    		    			me.startImageRetrieval(albumCoverData, me._coverSize);
				            return;
				        }
    				} catch (err) {
	    			    me.printLog("Failed to cache Item: " + err);
		    		}
			    }
			    if (albumCoverData) {
	    			// we have album data but could not download it to cache. just return it
    				me.callResponseHandler(albumCoverData);
		    		return;
			    }
			  }
    		} /* 200 response */
            else {
    			    me.printLog("Wrong server status" + status);
                    me.callResponseHandler(null);
                    return;
    		}

            // in case of any errors or failures ( except for  abort), try to get the next result
	    	if (!me._isAbortCalled) {
		        me.getNextAmazonResult();
		    } else {
    		    me.callResponseHandler(null);
    		}
    		} /* ready state 4 */
    	}
    	this._requestObject.send(null);
    	return;
    },

    parseResponse: function() {
        var albumCoverData = null;
	    try {
	        var doc = this._requestObject.responseXML;
            itemElements = doc.getElementsByTagName("Item");
    	    if (itemElements != null) {
    	    	albumCoverData = this.findMostReleventItem(itemElements)
	        } else {
		        this.trace("Failed to find Items");
    	    }
	    } catch (err) {
    	    this.printLog("Failed to find Items: " + err);
	    }  
        return albumCoverData;
    },
    
    abort: function() {
        if (this._isAbortCalled) return;

    	this._isAbortCalled = true;
        try {
        	if (this._requestObject) {
	   			this._requestObject.abort();
			}
        } catch(e) {
    	    this.printLog("Problems aborting xmlhttprequest: " + e);
        }

        try { 
			if (this._imageRetrievalChannel) {
	    		this._imageRetrievalChannel.cancel(Components.results.NS_ERROR_ABORT);
			}
        } catch(e) {
    	    this.printLog("Problems aborting image retreival: " + e);
        }
    },

    getAlbumCoverDetailsBySize : function(itemElement, size) {
        var albumCoverData = new AlbumCoverData(null, null, null, null, null);

        var albumCoverSize = itemElement.getElementsByTagName(size + "Image")[0]
        if (albumCoverSize) {
            var albumCoverUrlNode = albumCoverSize.getElementsByTagName("URL")[0];
            if (albumCoverUrlNode) {
                albumCoverData.imagePath = albumCoverUrlNode.firstChild.nodeValue;
            }
            var albumCoverHeightNode = albumCoverSize.getElementsByTagName("Height")[0];
            if (albumCoverHeightNode) {
                albumCoverData.height = albumCoverHeightNode.firstChild.nodeValue;
            }
            var albumCoverWidthNode = albumCoverSize.getElementsByTagName("Width")[0];
            if (albumCoverWidthNode) {
                albumCoverData.width = albumCoverWidthNode.firstChild.nodeValue;
            } 
        }

        return albumCoverData;
    },

    findMostReleventItem: function(itemElements) {
	    var bCheckTitle = false;
	    var bCheckTrack = false;
    	var album = this._album.toLowerCase();
    	var artist = this._artist.toLowerCase();

    	if (!album) {
    	    bCheckTrack = true;	
    	} else if ( (artist == album) || (artist == "various artists") ) {
	        bCheckTitle = true;
    	}
    
    	var albumCoverData = null;
    	var mostRelevanceAlbumCoverData = null;
    	var mostRelevanceAlbumTitle = null;

    	for (var i=0; i < itemElements.length; i++) {
    	    var currentItemElement = itemElements[i];
    	    var normTitle = null;

    	    try {
                var tempAlbumCoverData = this.getAlbumCoverDetailsBySize(currentItemElement, this._coverSize);
                var imageAttributes = currentItemElement.getElementsByTagName("ItemAttributes")[0];
                if (imageAttributes) {
                    var imageAttributesTitle = imageAttributes.getElementsByTagName("Title")[0];
                    if (imageAttributesTitle) {
                 		tempAlbumCoverData.title = imageAttributesTitle.firstChild.nodeValue;
                        normTitle = tempAlbumCoverData.title.toLowerCase();
                		this.trace("Album:  " + tempAlbumCoverData.title);
                    }
                }
            
                if ((this._coverSize == "Small") &&
				    ((tempAlbumCoverData.height < 70) || (tempAlbumCoverData.width < 70))) {
                    // look for medium image
                    var tryAlbumCoverData = this.getAlbumCoverDetailsBySize(currentItemElement, "Medium");
                    if (tryAlbumCoverData.imagePath && tryAlbumCoverData.height && tryAlbumCoverData.width) {
                        tempAlbumCoverData.imagePath = tryAlbumCoverData.imagePath;
                        tempAlbumCoverData.height = tryAlbumCoverData.height;
                        tempAlbumCoverData.width = tryAlbumCoverData.width;
                    }
                }

                var detailPageURLNode = currentItemElement.getElementsByTagName("DetailPageURL")[0];
                if (detailPageURLNode) {
        		    tempAlbumCoverData.link = detailPageURLNode.firstChild.nodeValue;
                }

                if (bCheckTrack) {
                    var tracksNode = currentItemElement.getElementsByTagName("Tracks")[0];
                    if (!tracksNode) {
                        continue;
                    }
                    var discList = tracksNode.getElementsByTagName("Disc");
    		        if (!this.isTrackExists(discList)) {
	    		            continue;
		            }
	            }

  	        } catch (err) {
    		    this.trace("Failed to find an attribute in item: " + i + " - " + err);
        		continue;
    	    }
    
    	    if (!tempAlbumCoverData.title || !tempAlbumCoverData.link || !tempAlbumCoverData.height 
                || !tempAlbumCoverData.width || !tempAlbumCoverData.imagePath) {
    	    	continue;
    	    }
    
    	    if (bCheckTitle) {
    	        if (normTitle == album) {
    	            albumCoverData = tempAlbumCoverData;
        		    this._amazonAlbumTitle = normTitle;
        		    break
    	        } else {
    	            if(normTitle.indexOf(album) == 0 && mostRelevanceAlbumCoverData == null) {
            			mostRelevanceAlbumCoverData = tempAlbumCoverData;
	     			    mostRelevanceAlbumTitle = normTitle;
    		        }
        		    continue;
       	    	}
    	    } else {
           		albumCoverData = tempAlbumCoverData;
    			this._amazonAlbumTitle = normTitle;
    		   break;
    	    }
    	}
    
    	if (albumCoverData == null)  { 
    	    albumCoverData = mostRelevanceAlbumCoverData;
    	    this._amazonAlbumTitle = mostRelevanceAlbumTitle;
    	}
    	return albumCoverData;
    },

    isTrackExists: function(discList) {
	    var normTrack = this.normalizeTitle(this._track);
    	if (discList == null) { 
	        return false;
    	}

    	for(var discIndex=0; discIndex < discList.length; discIndex++) {
    	    var currentDiscElement = discList[discIndex];
    	    if (currentDiscElement == null) continue;

            var trackElements = currentDiscElement.getElementsByTagName("Track");
       		if (trackElements == null) continue;

       	    for(var trackIndex=0; trackIndex < trackElements.length; trackIndex++) {
                var currentTrackElement = trackElements[trackIndex];
     	    	if (currentTrackElement == null) continue;

                var amazonNormTrack = this.normalizeTitle(currentTrackElement.firstChild.nodeValue);
                this.trace("track =" + amazonNormTrack + "|\t <->" + normTrack + "|");
  		        if (amazonNormTrack == normTrack) {
   	            		return true;
       			}
            }
	    }

    	return false;
    },

    normalizeTitle : function(str) {
        // remove 'and' 
        str = str.replace(/ and /ig, '');
        str = str.replace(/^and /ig, '');
        str = str.replace(/ and$/ig, '');

        // remove 'the'
        str = str.replace(/^the /ig, '');
        str = str.replace(/ the$/ig, '');

    	// remove special chars
        // 0 - '/'
        str = str.replace(/[\u0000-\u002F]/g, '');
        // ':' - '@'
        str = str.replace(/[\u003A-\u0040]/g, '');
        // '[' - '`'
        str = str.replace(/[\u005B-\u0060]/g, '');
        // '{' - 'DEL'
        str = str.replace(/[\u007B-\u007F]/g, '');
        // remove control chars: http://www.unicode.org/charts/PDF/U0080.pdf
	    str = str.replace(/[\u0080-\u00BF]/g, ''); 
    	str = str.replace(/\u2019/g, ''); // remove (\' like) char. in decimal = 8217
	    
        // remove parens
		str = str.replace(/\([^\)]*\)/g, '');
        str = str.replace(/\[[^\]]*\]/g, '');
        str = str.replace(/[()]/g, '');
        str = str.replace(/[\[\])]/g, '');

    	//remove whitespace
	    str = str.replace(/ /g, '');

        return str.toLowerCase();
    },

    generateUrl: function(tld) {
    	var escapeArtist = this.acgEncodeUri(this._artist);
    	var url = this._amazonURL.replace(/%artist%/, escapeArtist);
        
    	if (this._album) {
	        var escapeAlbum = this.acgEncodeUri(this._album);
    	   url = url.replace(/%album-query%/, "&Title=" + escapeAlbum);
    	} else {
    	    url = url.replace(/%album-query%/, "");
    	}
    	url = url.replace(/%tld%/, tld);
    	url = url.replace(/%amazon_key%/, this._amazonKey);
    	url = url.replace(/%amazon_tag%/, this._amazonTagArr[tld]);
	    var responseGroup = "";
    	if (!this._album) {
	        var responseGroup = ",Tracks";
    	}
    	url = url.replace(/%additional-response-group%/, responseGroup);
        return url;
    },

    getCoverFromCache: function(imageSize) {
	    if(!this._cacheManager) {
    	    return null;
    	}

	    var name = this._album;
        var bSearchUsingTrack = false;
    	if (!this._album) {
	        name = this._track;
            bSearchUsingTrack = true;
    	}

    	return this._cacheManager.get(name, this._artist, imageSize, bSearchUsingTrack);
    },
    
    startImageRetrieval: function(albumCoverData, imageSize) {
        var url = albumCoverData.imagePath;
	    var me = this;
    	var listener = {
	        onDataAvailable : function(request, context, aInputStream, offset, count) {
    		    inputStream.setInputStream(aInputStream);
    	    	bytes += inputStream.readBytes(count);
	        },

	        onStartRequest : function(request, context) {
		        inputStream = Components.classes['@mozilla.org/binaryinputstream;1'].
                            createInstance(Components.interfaces.nsIBinaryInputStream);
    	    },

	        onStopRequest : function(request, context, statusCode) {
		        var httpChannel = channel;
			var succeeded = null;
			try {
			    succeeded = httpChannel.requestSucceeded;
			} catch (e) {}
            		if (httpChannel && 
		                  succeeded && 
            			  Components.isSuccessCode(statusCode) &&
			              (bytes.length > 0)) {
        		            var imageType = httpChannel.getResponseHeader('Content-Type');
    		                me.saveInCache(albumCoverData, bytes, imageType, imageSize);
			}
		    }
	    }

    	var inputStream = null; 
	    var bytes = "";
    	var ioService = Components.classes['@mozilla.org/network/io-service;1']
		    .getService(Components.interfaces.nsIIOService);
	    var channel = ioService.newChannel(url, null, null)
                .QueryInterface(Components.interfaces.nsIHttpChannel);
	this._imageRetrievalChannel = channel;
    	channel.asyncOpen(listener, null);
    },

    saveInCache: function(albumCoverData, bytes, imageType, imageSize) {
    	var album = this._album;
	    var track = null;
    	if (!this._album) {
	        // when I dont have album from the player I am caching track title and album name that I found on amazon
	        album = this._amazonAlbumTitle;
    	    track = this._track;
	    }
    	var daysToExpire = this._daysToExpire;
	    if (bytes == null) {
	        daysToExpire = this._daysToExpireForNotFoundItem;
    	}
    
    	try {
	    	var cachedAlbumCoverData =  this._cacheManager.set(album, this._artist, track, imageSize, albumCoverData,
		    		     daysToExpire, bytes, imageType);
    	} catch (err) {
	        this.printLog("Failed to cache Item: " + err);
	        this.callResponseHandler(albumCoverData)
    	    return;
	    }
    	this.callResponseHandler(cachedAlbumCoverData);	
    },

    cacheNotFoundResult : function() {
	    var dummyAlbumCoverData = new AlbumCoverData(null, null, null, null);
    	this.saveInCache(dummyAlbumCoverData, null, null, this._coverSize);
    },

    callResponseHandler: function(albumCoverData) {
        if (albumCoverData && !albumCoverData.imagePath) {
            // we have data but no imagePath, meaning we cached 'non found' result
            albumCoverData = null;
        }

    	if (this._handler != null) {
	    	this._handler.handle(albumCoverData);
    	} else {
	        this.printLog("No handler!!!");
    	}
    },

    setTlds: function() {
    	var tld = null;
        var fallbackTLD = null;
    	switch (this._language) {
    	    case ENGLISH_LANG:
    	    case WESTERN_EUROPEAN_LANG:
    	    case UNKNOWN_LANG:
    	        switch (this._userCountry) {
       		        case USA_COUNTRY:
            			tld = 'com';
              			break;
		            case FRANCE_COUNTRY:
                   		tld = 'fr';
            			fallbackTLD = 'com';
            			break;
		           case CANADA_COUNTRY:
            			tld = 'ca';
            			fallbackTLD = 'com';
            			break;
            	   case GERMANY_COUNTRY:
            		    case AUSTRIA_COUNRTY:
            			tld = 'de';
            			fallbackTLD = 'com';
            			break;
          		    case BRITAIN_COUNTRY:
            			tld = 'co.uk';
            			fallbackTLD = 'com';
            			break;
        		    case JAPAN_COUNTRY:
		            	tld = 'co.jp';
            			fallbackTLD = 'com';
            			break;
        		    default:
		            	tld = 'com';
            			break;
    		    }
	    	break;
    
    	    case FRENCH_LANG: 
	            switch (this._userCountry) {
	        	    case CANADA_COUNTRY:
		            case USA_COUNTRY:
            			tld = 'ca';
            			fallbackTLD = 'fr';
            			break;
                    default:
            			tld = 'fr';
            			fallbackTLD = 'com';
            			break;
        		}
		        break;
           case GERMAN_LANG: 
                tld = 'de';
            	fallbackTLD = 'com';
            	break;
           case EAST_ASIAN_LANG:
    	   case JAPANESE_LANG: 
            	tld = 'co.jp';
        		break; 
    	    default:
	        	this.printLog("Ilegal language: " + this._language);
	    }

        if (tld) {
            this._tlds.push(tld);
        }

        if (fallbackTLD) {
            this._tlds.push(fallbackTLD);
        }
    }, 

    acgEncodeUri: function(str) {
        if(this._currentTld == 'co.jp') {
            return encodeURIComponent(str); 
        }
        //Currently the language for all other tld  will be western 
        return escape(str);
    },

    printLog : function(str) {
        dump(str + '\n');
    },

    trace : function(message) {
    //    dump(message + '\n');
    }
}


// Class: AlbumCoverData
function AlbumCoverData(imagePath, link, height, width, title) {
    this.imagePath = imagePath;
    this.link = link;
    this.height = height;
    this.width = width;
    this.title = title;
}
